#include "midi.h"
#include <stdio.h>

#define MAXINT		0x7fffffff

struct Dest : public vector<unsigned short> {
	void push_note(int voice, int note, int velo) {
		vector<unsigned short>::push_back(voice << 11 | velo >> 3 << 7 | note + 1);
	}
	void push_wait(int wait) {
		vector<unsigned short>::push_back(-wait);
	}
};

static Dest dest;

class Voice {
public:
	enum { N = 8 };
	Voice() { channel = note = velo = req = 0; elapse = MAXINT; }
	static void NoteOn(int channel, int note, int velo) {
		int max = 0;
		Voice *p, *maxp = voice;
		for (p = voice; p < voice + N; p++) {
			if (p->elapse == MAXINT) {
				maxp = p;
				break;
			}
			int v = 128 - p->velo;
			if (max < v) max = v, maxp = p;
		}
		if (maxp->elapse < MAXINT) maxp->req = 0, offcount++;
		if (maxvoice < maxp - voice) maxvoice = maxp - voice;
		maxp->On(channel, note, velo);
	}
	static void NoteOff(int channel, int note) {
		Voice *p;
		for (p = voice; p < voice + N && 
			(!(p->channel & 1 << channel) || p->note != note); p++) 
			;
		if (p < voice + N && !(p->channel &= ~(1 << channel))) p->Off();
	}
	static void Dump(int duration) {
		Voice *p;
		for (p = voice; p < voice + N; p++) {
			if (p->req) {
				p->req = 0;
				dest.push_note(p - voice, p->note, p->velo);
			}
			if (p->elapse < MAXINT) p->elapse += duration;
		}
	}
	static int OffCount() { return offcount; }
	static int MaxVoice() { return maxvoice + 1; }
private:
	void On(int _channel, int _note, int _velo) {
		channel = 1 << _channel;
		note = _note;
		velo = _velo;
		elapse = 0;
		req = 1;
	}
	void Off() {
		req = 1;
		channel = note = velo = 0;
		elapse = MAXINT;
	}
	int channel, note, velo, elapse, req;
	static Voice voice[N];
	static int offcount, maxvoice;
};

Voice Voice::voice[Voice::N];
int Voice::offcount, Voice::maxvoice;
static int enable = 0xfdff, started, base, fs = 44100;

static void note(int channel, int note, int velo) {
	if (enable & 1 << channel) {
		note -= base;
		if (velo) Voice::NoteOn(channel, note, velo);
		else Voice::NoteOff(channel, note);
		started = 1;
	}
}

static void wait(int duration) {
	if (started) {
		duration = int(duration / (1e6 / fs) + 0.5);
		Voice::Dump(duration);
		while (duration > 0x8000) {
			dest.push_wait(0x8000);
			duration -= 0x8000;
		}
		dest.push_wait(duration);
	}
}

int main(int argc, char *argv[]) {
	while (argc > 2 && argv[1][0] == '-') {
		argc--;
		argv++;
		switch (argv[0][1]) {
			case 'e':
			sscanf(argv[1], "%x", &enable);
			argc--;
			argv++;
			break;
			case 't':
			sscanf(argv[1], "%d", &base);
			base = -base;
			argc--;
			argv++;
			break;
		}
	}
	if (argc != 2) {
		fprintf(stderr, "Usage: mid2pic [-e <enable channel>] [-t <transpose>] <filename>\n");
		return 1;
	}
	Midi midi;
	midi.Parse(argv[1]);
	int i, min, max;
	midi.GetNoteMinMax(min, max);
	if (min - base < 0) base += min - base;
	midi.Dump(note, wait);
	printf("MaxVoice: %d  Note: %d-%d\n", Voice::MaxVoice(), min - base, max - base);
	if (base) printf("Transpose: %d\n", -base);
	if (Voice::OffCount()) printf("OffCount: %d\n", Voice::OffCount());
	FILE *fo = fopen("orgel.dat", "wb");
	if (fo) {
		fwrite(&dest.front(), 2, dest.size(), fo);
		fclose(fo);
		printf("%d bytes\n", 2 * (int)dest.size());
	}
	return 0;
}

